module Main where

import Eval                   (evalTricu, mainResult, result)
import FileEval
import Parser                 (parseTricu)
import REPL
import Research
import ContentStore

import Control.Monad          (foldM)
import Control.Monad.IO.Class (liftIO)
import Data.Version           (showVersion)
import Text.Megaparsec        (runParser)
import Paths_tricu            (version)
import System.Console.CmdArgs

import qualified Data.Map as Map

data TricuArgs
  = Repl
  | Evaluate { file :: [FilePath], form :: EvaluatedForm }
  | TDecode { file :: [FilePath] }
  deriving (Show, Data, Typeable)

replMode :: TricuArgs
replMode = Repl
  &= help "Start interactive REPL"
  &= auto
  &= name "repl"

evaluateMode :: TricuArgs
evaluateMode = Evaluate
  { file = def &= help "Input file path(s) for evaluation.\n \
      \ Defaults to stdin."
      &= name "f" &= typ "FILE"
    , form = TreeCalculus &= typ "FORM"
      &= help "Optional output form: (tree|fsl|ast|ternary|ascii|decode).\n \
      \ Defaults to tricu-compatible `t` tree form."
      &= name "t"
  }
  &= help "Evaluate tricu and return the result of the final expression."
  &= explicit
  &= name "eval"

decodeMode :: TricuArgs
decodeMode = TDecode
  { file = def
    &= help "Optional input file path to attempt decoding.\n \
    \ Defaults to stdin."
    &= name "f" &= typ "FILE"
  }
  &= help "Decode a Tree Calculus value into a string representation."
  &= explicit
  &= name "decode"

main :: IO ()
main = do
  let versionStr = "tricu Evaluator and REPL " ++ showVersion version
  args <- cmdArgs $ modes [replMode, evaluateMode, decodeMode]
    &= help "tricu: Exploring Tree Calculus"
    &= program "tricu"
    &= summary versionStr
    &= versionArg [explicit, name "version", summary versionStr]
  case args of
    Repl -> do
      putStrLn "Welcome to the tricu REPL"
      putStrLn "You may exit with `CTRL+D` or the `!exit` command."
      putStrLn "Try typing `!` with tab completion for more commands."
      repl
    Evaluate { file = filePaths, form = form } -> do
      result <- case filePaths of
        [] -> runTricuT <$> getContents
        (filePath:restFilePaths) -> do
            initialEnv <- evaluateFile filePath
            finalEnv <- foldM evaluateFileWithContext initialEnv restFilePaths
            pure $ mainResult finalEnv
      let fRes = formatT form result
      putStr fRes
    TDecode { file = filePaths } -> do
      value <- case filePaths of
        [] -> getContents
        (filePath:_) -> readFile filePath
      putStrLn $ decodeResult $ result $ evalTricu Map.empty $ parseTricu value

runTricu :: String -> String
runTricu = formatT TreeCalculus . runTricuT

runTricuT :: String -> T
runTricuT input =
  let asts     = parseTricu input
      finalEnv = evalTricu Map.empty asts
   in result finalEnv

runTricuEnv :: Env -> String -> String
runTricuEnv env = formatT TreeCalculus . runTricuTEnv env

runTricuTEnv :: Env -> String -> T
runTricuTEnv env input =
  let asts     = parseTricu input
      finalEnv = evalTricu env asts
   in result finalEnv

runTricuWithEnvT :: String -> (Env, T)
runTricuWithEnvT input =
  let asts     = parseTricu input
      finalEnv = evalTricu Map.empty asts
   in (finalEnv, result finalEnv)

runTricuWithEnv :: String -> (Env, String)
runTricuWithEnv input =
  let asts     = parseTricu input
      finalEnv = evalTricu Map.empty asts
      res      = result finalEnv
   in (finalEnv, formatT TreeCalculus res)

runTricuEnvWithEnvT :: Env -> String -> (Env, T)
runTricuEnvWithEnvT env input =
  let asts     = parseTricu input
      finalEnv = evalTricu env asts
   in (finalEnv, result finalEnv)

runTricuEnvWithEnv :: Env -> String -> (Env, String)
runTricuEnvWithEnv env input =
  let asts     = parseTricu input
      finalEnv = evalTricu env asts
      res      = result finalEnv
   in (finalEnv, formatT TreeCalculus res)
